// Dice
// Roll a pair of dice with the ATtiny13A
// Using counter/timer interrupt for refresh

// LEDs for dice are wired as follows:
// A1    B1    A2    B2
// C1 D1 E1    C2 D2 E2
// F1    G1    F2    G2
//
// Each charlieplexed load is one series pair of LEDs which
// correspond to two opposing dots that are always lit together.
// The exception is D1 and D2, which are each a single LED in series
// with a 220 ohm resistor to reduce its brightness to match the others.
//
// PB0-PB3 are used for charlieplexing as follows:
//
// LED	Anode	Cathode	Mux Bit
// ===	=====	=======	=======
// AG1	 PB0		PB1				0
// BF1	 PB1		PB0				1
// CE1	 PB0		PB2				2
// D1		 PB2		PB0				3
// AG2	 PB1		PB3				4
// BF2	 PB3		PB1				5
// CE2	 PB2		PB3				6
// D2		 PB3		PB2				7
//
// The actual multiplexing is done in the timer's interrupt service routine.
//
// PB4 is a pushbutton to ground (to "roll" the dice), using
// the processor's internal 10k pullup.
//
// Author:	Bruce R. Ratoff (ko4xl@yahoo.com)
// Date:		April 10, 2012
//
// This code may be freely distributed with or without modification, only if
// credit is given to the original author and all changes are noted below.
//
// Date				Author		Change
// ====				======		======
// 04/22/2012	B Ratoff	Enter power down state if idle for 1 minute
//											(Wake up on next button press)

//Include Files

#include <avr/io.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/power.h>
#include <avr/sleep.h>
#include <avr/wdt.h>

#define LED_TIMEOUT 30000U
#define PWR_TIMEOUT 30000U

// DDR bits for each LED pair
// Bit 4 is always input (for the button)
static const char ddrlist[] PROGMEM =
	{0b00000011, 0b00000011, 0b00000101, 0b00000101, 0b00001010, 0b00001010, 0b00001100, 0b00001100};

// Output bits for each LED pair
// Bit 4 is always set to enable button pullup
static const char portlist[] PROGMEM =
	{0b00000001, 0b00000010, 0b00000001, 0b00000100, 0b00000010, 0b00001000, 0b00000100, 0b00001000};
	
// Patterns for each die side - bit 0 1 2 3 represents AG BF CE D
// Second die is the same as first but shifted to high nibble
static const char dielist[] PROGMEM =
	{0b00001000, 0b00000001, 0b00001001, 0b00000011, 0b00001011, 0b00000111,
	 0b10000000, 0b00010000, 0b10010000, 0b00110000, 0b10110000, 0b01110000};

volatile unsigned char lights_on;			// Pattern of lights that is being displayed
volatile unsigned int time_on;				// Milliseconds lights have been on
volatile unsigned int time_off;				// Milliseconds power has been on with lights off and no button push

volatile char die1 = 0;								// Current value for first die (will take values 0-5)
volatile char die2 = 6;								// Current value for second die (will take values 6-11)

// Timer interrupt service routine
// This interrupt fires once per millisecond.  It is responsible for three things:
// 1 - Perform multiplexing of current light pattern
// 2 - Keep count of how long lights have been on
// 3 - Keep count of how long power has been on with no lights or button push
//
ISR(TIM0_COMPA_vect)
{
	static char l=0;			// Count of which bit in light pattern we're working on
	static char m=1;			// Bit mask for the bit we're working on
	
	DDRB &= 0xF0;					// Switch all I/O bits to input, turning off all lights
	PORTB &= 0xF0;				// Disable internal pullup resistors for all I/O bits
	
	// If the current bit in the pattern is set (i.e. light should be lit),
	// 1 - Look up the corresponding I/O bits and set their direction to output
	// 2 - Look up and set the output bit pattern to turn on the correct lights
	if(lights_on & m) {
		DDRB |= pgm_read_byte(&ddrlist[l]) & 0x0F;
		PORTB |= pgm_read_byte(&portlist[l]) & 0x0F;
	}
	m <<= 1;							// Advance bit mask to next bit in pattern
	if(++l > 7) {				// Cycle back to beginning after all 8 bits have been processed
		m = 1;
		l = 0;
	}
	if( lights_on ) {									// Keep track of how long lights have been lit
		if( time_on < LED_TIMEOUT+1 )
			++time_on;
	} else {													// Keep track of how long power has been on with no lights or button
		if( time_off < PWR_TIMEOUT+1 )
			++time_off;
	}
}

// Pin change interrupt service routine
// Fires when button is pushed while processor is in power down state
// Since the interrupt itself turns the processor back on, all we have to do is clear the idle time counter
ISR(PCINT0_vect)
{
	time_on = 0;
}

// Main program
int main(void) {
	char t,i,j;
	
	// Start by initializing everything
	i = 0;
	
	wdt_disable();				// Kill the watchdog timer to save power since we don't use it
	power_adc_disable();	// Kill the ADC for the same reasons
	PRR |= PRADC;
	set_sleep_mode(SLEEP_MODE_PWR_DOWN);		// Set sleep mode to full power down
	
	DDRB =  0b00000000;		// Set all bits to input
	PORTB = 0b00010000;		// Enable button pullup
	
	// Set up counter/timer
	TCCR0A = 0<<COM0A1 | 0<<COM0A0 | 0<<COM0B1 | 0<<COM0B0 | 1<<WGM01 | 0<<WGM00;	// Upper limit is OCR0A
	TCCR0B = 0<<FOC0A | 0<<FOC0B | 0<<WGM02 | 0<<CS02 | 1<<CS01 | 0<<CS00;				// 1.2MHz / 8 = 150KHz
	OCR0A = 150;															// Divide down to 1ms period
	TIFR0 = 1<<OCF0B | 1<<OCF0A | 1<<TOV0;		// Clear all timer interrupt flags
	TIMSK0 = 0<<OCIE0B | 1<<OCIE0A | TOIE0;	// Enable interrupt on timer match
	
	GIMSK = 0<<INT0 | 0<<PCIE;		// Start out with pin interrupt sources disabled
	PCMSK = 1<<PCINT4;						// Select pin change interrupt on button pin
	
	sei();		// This should start the 1ms timer ticking
	
	// Main program loop
	// Check for timeouts to kill lights or power down
	// Keep dice rolling until button is pushed and released, then
	// 	display the new dice pattern (actually just sets it up for the timer interrupt to display)
	while(1) {
		// Wait for button to be pushed
	  while(PINB & 0b00010000) {
			// If idle long enough, turn off lights to save power
			if( lights_on ) {
				if(time_on > LED_TIMEOUT-1) {
					lights_on = 0;
					time_off = 0;
				}
			} else {
				// If lights off long enough, power down (pin change will wake us)
				cli();
				if(time_off > PWR_TIMEOUT-1) {
					GIFR |= 1<<PCIF;		// clear pin change flag
					GIMSK |= 1<<PCIE;		// Wake on pin change (button pushed)
					sleep_enable();
					sei();
					sleep_cpu();
					sleep_disable();
					GIMSK &= ~(1<<PCIE);
				}
				sei();
			}
			// no button or timeout yet - spin the first die while we are waiting
			if(++die1 > 5)
				die1 = 0;
		}
		// We get here once the button is pushed
		lights_on = 0;		// turn off old dice
		_delay_ms(20);		// debounce the button press
		// wait for button to be released
	  while(!(PINB & 0b00010000)) {
			// spin the second die while we're waiting
			if(++die2 > 11)
				die2 = 6;
		}
		_delay_ms(20);		// debounce button release
		
		// Animate the lights in rolling pattern, gradually slowing down
		for(t=1; t<60; ++t) {
			lights_on = pgm_read_byte(&dielist[i]) | pgm_read_byte(&dielist[11-i]);
			for(j=0; j<t; ++j)
				_delay_ms(2);
			if(++i > 11)
				i = 0;
		}
		
		// Now display the light pattern for the actual roll
		lights_on = pgm_read_byte(&dielist[die1]) | pgm_read_byte(&dielist[die2]);
		time_on = 0;			// Clear on-time counter
	}
}
